"""Core module."""
from __future__ import annotations

import pandas as pd

from auto_smooth import savgol

TYPES = ["savgol"]


def main(
    data: pd.Series | pd.DataFrame,
    kind: str | None = None,
    verbose: int = 0,
) -> pd.Series | pd.DataFrame:
    if kind is None:
        kind = "savgol"

    if kind == "savgol":
        data_filtered = savgol.auto_savgol(data, verbose=verbose)
    else:
        raise NotImplementedError(
            "Only the following smoothing/filtering techniques are implemented:"
            f"{TYPES}"
        )

    return data_filtered



import pandas as pd

from auto_smooth import utils


def test_generate_series():
    num_periods = 100
    seed = 1
    data = utils.generate_series(num_periods, seed)

    assert isinstance(data, pd.Series)
    assert data.size == num_periods
    assert int(data.iloc[-1]) == 92



"""Plotting module."""
from __future__ import annotations

import matplotlib.pyplot as plt
import pandas as pd

# set style
plt.style.use("fivethirtyeight")
plt.rcParams["lines.linewidth"] = 3
plt.rcParams["font.size"] = 10
plt.rcParams["figure.figsize"] = (6, 3)


def smooth(
    data: pd.DataFrame | pd.Series,
    data_filtered: pd.DataFrame | pd.Series,
    ax=None,
) -> None:
    """Plot smooth data along with the original."""
    ax = data.plot(alpha=0.7, label="Original", ax=ax)
    data_filtered.plot(ax=ax, label="Smooth")

    ax.xaxis.grid(False)
    ax.set_title("Original vs. Smooth", fontsize=12)
    ax.legend(fontsize=8)


def residuals(
    data: pd.DataFrame | pd.Series,
    data_filtered: pd.DataFrame | pd.Series,
    ax=None,
) -> None:
    """Plot residuals."""
    ax = (data - data_filtered).plot(style="o", alpha=0.7, ax=ax)
    ax.xaxis.grid(False)

    ax.axhline(y=0, linestyle="-", linewidth=2)
    ax.set_title("Residuals", fontsize=12)


def subplots(
    data: pd.DataFrame | pd.Series,
    data_filtered: pd.DataFrame | pd.Series,
) -> None:
    """Plot smoothing and residuals in subplots."""
    fig, ax = plt.subplots(nrows=2, ncols=1, gridspec_kw={"height_ratios": [3, 2]})
    fig.set_size_inches(6, 6)
    fig.tight_layout(pad=4.0)

    smooth(data, data_filtered, ax=ax[0])
    residuals(data, data_filtered, ax=ax[1])
    plt.show()



"""Metrics module.

This module allows user to evaluated the filtering results.
"""
from __future__ import annotations

import numpy as np
import pandas as pd


def mse(y_true: pd.Series, y_hat: pd.Series) -> pd.Series:
    """Mean squared error."""
    return ((y_true - y_hat) ** 2).mean()


def rmse(y_true: pd.Series, y_hat: pd.Series) -> pd.Series:
    """Root mean squared error."""
    return np.sqrt(mse(y_true, y_hat))


def mae(y_true: pd.Series, y_hat: pd.Series) -> pd.Series:
    """Mean absolute error."""
    return np.mean(np.abs(y_true - y_hat))


def r2(y_true: pd.Series, y_hat: pd.Series) -> pd.Series:
    "R-squared."
    ss_tot = ((y_true - y_true.mean()) ** 2).sum()
    ss_res = ((y_true - y_hat) ** 2).sum()
    return 1 - (ss_res / ss_tot)


def get_scores(y_true: pd.Series, y_hat: pd.Series) -> dict[str, pd.Series]:
    """Get all scores."""
    scores = {
        "rmse": rmse(y_true, y_hat),
        "mae": mae(y_true, y_hat),
        "r2": r2(y_true, y_hat),
    }

    return scores



"""Core module."""
from __future__ import annotations

import numpy as np
import pandas as pd
from scipy.signal import savgol_filter

from auto_smooth import metrics, plotting


def savgol(
    data: pd.Series,
    window_length: int,
    polyorder: int,
    plot: bool = True,
    **kwargs,
) -> pd.Series:
    """Apply savgol filter.

    Parameters
    ----------
    window_length:
        Window size.
    polyorder:
        Order of the polinomial.
    plot:
        Whether to plot the comparison of the original
        data and the filtered data.
    kwargs:
        Keyword arguments to be passed to the scipy
        savgol_filter function.

    Returns
    -------
    df_filtered

    """
    x = data.dropna()

    # apply savgol filter
    y = savgol_filter(x=x, window_length=window_length, polyorder=polyorder, **kwargs)

    # convert y to Series
    data_filtered = pd.Series(y, index=x.index, name=f"{data.name}_filtered")

    # reindex back with the original index
    data_filtered = data_filtered.reindex(data.index)

    if plot:
        plotting.subplots(data, data_filtered)
        # plotting.residuals(data, data_filtered)

    return data_filtered


def auto_savgol(
    data: pd.Series,
    wl_min: int | None = None,
    wl_max: int | None = None,
    po_min: int | None = None,
    po_max: int | None = None,
    max_samples: int = 50,
    metric: str = "rmse",
    plot: bool = True,
    verbose: int = 0,
) -> pd.Series:
    """Perform auto-savgol to detect best wl/po values.

    Parameters
    ----------
    data:
        Pandas Series.
    wl_min:
        Minimum window length. If None, wl_min will be
        automatically selected.
    wl_max:
        Maximum window length.
    po_min:
        Minimum value of the polyorder.
    po_max:
        Maximum value of the polyorder.
    max_samples:
        Maximum number of samples between wl_min/po_min
        and wl_max/po_max to generate.
    metric: {"rmse", "mae", "r2"}
        Metric to choose the best wl/po parameters.
    verbose:
        Whether to print out results.
        - 0: silent
        - 1: best results
        - 2: all results

    Returns
    -------
    data_filtered

    """
    max_wl_po_ratio = 3

    if po_min is None:
        po_min = 2

    if po_max is None:
        po_max = 10

    if wl_min is None:
        wl_min = int(po_min * max_wl_po_ratio)

    if wl_max is None:
        wl_max = int(np.sqrt(data.size) * 2)

    # create an array of numbers spaced evenly
    wl_grid = np.unique(np.linspace(wl_min, wl_max, max_samples).astype(int))
    # on a log scale (a geometric progression)
    # wl_grid = np.unique(np.geomspace(wl_min, wl_max).astype(int))

    results: dict = {}

    for wl in wl_grid:
        for po in range(po_min, po_max + 1):
            # make sure the po is significantly lower than coefficients
            if wl / po < max_wl_po_ratio:
                continue

            # filtered data
            data_filtered = savgol(data, window_length=wl, polyorder=po, plot=False)

            # get scores
            result = metrics.get_scores(y_true=data, y_hat=data_filtered)

            if verbose == 2:
                print(f"{wl=:2}, {po=:2} >> {result}")

            results[(wl, po)] = result

    # create a DatFrame of scores for each (wl,po) pair
    df_scores = pd.DataFrame(results)
    # get the best (minimum scored) (wl,po) pair
    wl_best, po_best = df_scores.loc[metric].idxmin()

    if verbose:
        print(f"{wl_best=}, {po_best=}")

    data_filtered = savgol(data, window_length=wl_best, polyorder=po_best, plot=plot)

    return data_filtered



"""Utilities module."""
from __future__ import annotations

import numpy as np
import pandas as pd


def generate_series(
    num_periods: int = 100,
    seed: int | None = None,
) -> pd.Series:
    """Generate time series."""
    # Generate random steps with mean=0 and std=1
    rng = np.random.default_rng(seed=seed)
    random_steps = rng.normal(loc=0, scale=1.0, size=num_periods)
    # legacy
    # random_steps = np.random.normal(loc=0, scale=1.0, size=num_periods)

    # Simulate stock prices, with a starting price of 100
    random_steps[0] = 0
    price = 100 + np.cumsum(random_steps)

    # convert to series with datetime index
    data = pd.Series(
        price, index=pd.date_range(start="2022", periods=num_periods), name="data"
    )

    return data
